<?php

namespace Drupal\views_csv_source\Query;

use Drupal\Core\Database\Query\ConditionInterface;
use Drupal\Core\DependencyInjection\DependencySerializationTrait;
use League\Csv\Reader;
use League\Csv\Statement;
use League\Csv\TabularDataReader;

/**
 * Defines a class for building and executing a select query.
 */
class Select {

  use DependencySerializationTrait;

  /**
   * The direction for ascending order.
   */
  const DIRECTION_ASC = 'ASC';

  /**
   * The direction for descending order.
   */
  const DIRECTION_DESC = 'DESC';

  /**
   * The JSON representation of the query.
   */
  protected string $json = '{}';

  /**
   * The query metadata for alter purposes.
   */
  public array $alterMetaData;

  /**
   * Indicates if preExecute() has already been called.
   */
  protected bool $prepared = FALSE;

  /**
   * The connection to use.
   */
  protected array $records;

  /**
   * The grouped and aggregated records.
   */
  protected array $groupedAndAggregatedRecords = [];

  /**
   * Select constructor.
   *
   * @param \Drupal\views_csv_source\Query\Connection $connection
   *   The connection to use.
   * @param array $queryOptions
   *   The query options.
   */
  public function __construct(protected Connection $connection, protected array $queryOptions) {
  }

  /**
   * Check if the query is prepared.
   *
   * @return bool
   *   TRUE if the query is prepared, FALSE otherwise.
   */
  public function isPrepared(): bool {
    return $this->prepared;
  }

  /**
   * Prepare the query.
   *
   * @param \Drupal\views_csv_source\Query\Select|null $query
   *   The query to prepare.
   *
   * @return bool
   *   TRUE if the query is prepared, FALSE otherwise.
   */
  public function preExecute(Select $query = NULL): bool {
    if (!isset($query)) {
      $query = $this;
    }

    // Only execute this once.
    if ($query->isPrepared()) {
      return TRUE;
    }

    $this->prepared = TRUE;
    return $this->prepared;
  }

  /**
   * Execute the query.
   *
   * @return array|\Iterator
   *   The records.
   *
   * @throws \League\Csv\Exception
   *   If the CSV reader cannot be built.
   */
  public function execute(): array|\Iterator {
    return $this->getRecords();
  }

  /**
   * Fetch a field from the query.
   *
   * @return array|\Iterator|int
   *   The field.
   *
   * @throws \League\Csv\Exception
   *   If the CSV reader cannot be built.
   */
  public function fetchField(): array|\Iterator|int {
    $records = $this->getRecords();
    if ($this->isCountQuery()) {
      return count($records);
    }

    return $records;
  }

  /**
   * Check if the query is a count query.
   *
   * @return bool
   *   TRUE if the query is a count query, FALSE otherwise.
   */
  public function isCountQuery(): bool {
    $is_count = FALSE;
    if ($count_item = $this->getObjectItem('count')) {
      $is_count = $count_item == 'true' || $count_item == 1;
    }
    return $is_count;
  }

  /**
   * Add a field to the query.
   *
   * @param string $field
   *   The field to add.
   * @param string $function
   *   The function to apply to the field.
   *
   * @return $this
   *   The current object.
   */
  public function addField(string $field, string $function = ''): static {
    $column = new \stdClass();
    $column->name = $field;
    $column->function = $function;
    $this->addColumn($column);
    return $this;
  }

  /**
   * Add a group by to the query.
   *
   * @param string $field
   *   The field to group by.
   *
   * @return $this
   *   The current object.
   */
  public function groupBy(string $field): static {
    $group_by = $this->getObjectItem('group_by');
    if (!in_array($field, $group_by)) {
      $group_by[] = $field;
    }
    return $this->setObjectItem('group_by', $group_by);
  }

  /**
   * Add an order by to the query.
   *
   * @param string $column
   *   The column to order by.
   * @param string $direction
   *   The direction to order by.
   *
   * @return $this
   *   The current object.
   */
  public function orderBy(string $column, string $direction = self::DIRECTION_ASC): static {
    $data = $this->getObjectItem('order_by');
    $order_by = new \stdClass();
    $order_by->column = $column;
    $order_by->direction = $direction;
    $data[] = $order_by;
    return $this->setObjectItem('order_by', $data);
  }

  /**
   * Set the range.
   *
   * @param int $offset
   *   The offset.
   * @param int $limit
   *   The limit.
   *
   * @return $this
   *   The current object.
   */
  public function range(int $offset, int $limit): static {
    $range = (object) $this->getObjectItem('range');
    $range->offset = $offset;
    $range->limit = $limit;
    return $this->setObjectItem('range', $range);
  }

  /**
   * Add a tag to the query.
   *
   * @param string $tag
   *   The tag to add.
   *
   * @return $this
   *   The current object.
   */
  public function addTag(string $tag): static {
    $tags = $this->getObjectItem('tags');
    if (!in_array($tag, $tags)) {
      $tags[] = $tag;
    }
    return $this->setObjectItem('tags', $tags);
  }

  /**
   * Set the limit.
   *
   * @param int $limit
   *   The limit.
   *
   * @return $this
   *   The current object.
   */
  public function setLimit(int $limit): static {
    $range = (object) $this->getObjectItem('range');
    $range->limit = $limit;
    return $this->setObjectItem('range', $range);
  }

  /**
   * Set the offset.
   *
   * @param int $offset
   *   The offset.
   *
   * @return $this
   *   The current object.
   */
  public function setOffset(int $offset): static {
    $range = (object) $this->getObjectItem('range');
    $range->offset = $offset;
    return $this->setObjectItem('range', $range);
  }

  /**
   * Get the limit.
   *
   * @return int
   *   The limit.
   */
  public function getLimit(): int {
    $range = (object) $this->getObjectItem('range');
    return $range->limit ?? -1;
  }

  /**
   * Get the offset.
   *
   * @return int
   *   The offset.
   */
  public function getOffset(): int {
    $range = (object) $this->getObjectItem('range');
    return $range->offset ?? 0;
  }

  /**
   * Add a condition to the query.
   *
   * @param \Drupal\Core\Database\Query\ConditionInterface $condition
   *   The condition to add.
   *
   * @return $this
   *   The current object.
   */
  public function condition(ConditionInterface $condition): static {
    $where = [];
    $conditions = $condition->conditions();
    if (!empty($conditions['#conjunction'])) {
      $where['conjunction'] = $conditions['#conjunction'];
      unset($conditions['#conjunction']);
      $where['conditions'] = [];
      foreach ($conditions as $condition) {
        $where['conditions'][] = $this->buildConditions($condition);
      }
    }

    return $this->setObjectItem('where', $where);
  }

  /**
   * Add metadata to the query.
   *
   * @param string $key
   *   The key of the metadata to add.
   * @param mixed $object
   *   The metadata to add.
   *
   * @return $this
   *   The current object.
   */
  public function addMetaData(string $key, mixed $object): static {
    $this->alterMetaData[$key] = $object;
    return $this;
  }

  /**
   * Get the metadata.
   *
   * @param string $key
   *   The key of the metadata to get.
   *
   * @return mixed
   *   The metadata.
   */
  public function getMetaData(string $key): mixed {
    return $this->alterMetaData[$key] ?? NULL;
  }

  /**
   * Check if the query is empty.
   *
   * @return bool
   *   TRUE if the query is empty, FALSE otherwise.
   */
  public function isEmpty(): bool {
    $json = (array) $this->json;
    return empty($json);
  }

  /**
   * Get the count query.
   *
   * @return $this
   *   The count query.
   */
  public function countQuery(): static {
    $query = clone $this;
    $query->setCount();
    return $query;
  }

  /**
   * Set the count flag.
   *
   * @return $this
   *   The current object.
   */
  private function setCount(): static {
    return $this->setObjectItem('count', TRUE);
  }

  /**
   * Get the records.
   *
   * @return array|\Iterator
   *   The records.
   *
   * @throws \League\Csv\Exception
   *   If the CSV reader cannot be built.
   */
  private function getRecords(): array|\Iterator {
    $field_keys = array_keys($this->getSelectedFields());
    if (!$field_keys) {
      return [];
    }

    $options = $this->queryOptions += [
      'delimiter' => ',',
      'header_index' => 0,
    ];
    $uri = $this->connection->parseCsvUri();

    $csv = Reader::createFromString($this->connection->fetchContent($uri, $options));
    $csv->setDelimiter($options['delimiter']);
    if (isset($options['header_index']) && $options['header_index'] !== NULL) {
      $csv->setHeaderOffset($options['header_index']);
    }

    $stmt = Statement::create();

    // 1. Apply filters here.
    $stmt = $this->applyFilters($stmt);
    // Only select the needed fields.
    $csv = $csv->select(...$field_keys);

    // 2. Apply grouping here.
    if ($group_by = $this->getObjectItem('group_by')) {
      // Processing filtered csv to reduce the number of records to be grouped
      // if necessary.
      $records = $stmt->process($csv);
      if ($records->count() === 0) {
        return [];
      }

      $selected_aggregation_fields = array_filter($this->getSelectedFields(), fn(string $function) => $function && $function !== 'group_by');
      // Applying group by and aggregation callback.
      $records = $this->applyGroupBy($records, $group_by, $selected_aggregation_fields);

      // Initialize a new statement that doesn't have any of the filtering
      // closures.
      $group_by_stmt = Statement::create();
      // Process the grouping.
      $group_by_stmt->process($records);
    }

    // 3. Flatten the group by and aggregation array to build a new csv reader.
    if ($this->groupedAndAggregatedRecords) {
      $csv = $this->buildCsvReader($options);
    }

    // 4. Apply Sort.
    if (!$this->isCountQuery()) {
      // Only applying sorting for a non-counting query.
      $stmt = $this->applyOrderBy($stmt);
    }

    // 5. Apply range.
    $offset = $this->getOffset();
    $limit = $this->getLimit();
    if ($limit > 0) {
      $stmt = $stmt->limit($limit)->offset($offset);
    }

    $result_set = $stmt->process($csv);
    return $result_set->getRecords();
  }

  /**
   * Get the JSON representation of the query.
   *
   * @return string
   *   The JSON representation of the query.
   */
  public function __toString(): string {
    return $this->json;
  }

  /**
   * Apply the filters to the statement.
   *
   * @param \League\Csv\Statement $stmt
   *   The statement to apply the filters to.
   *
   * @return \League\Csv\Statement
   *   The statement with the filters applied.
   */
  protected function applyFilters(Statement $stmt): Statement {
    if ($where = $this->getObjectItem('where')) {
      return $stmt->where(function (array $record) use ($where): bool {
        return $record && static::conditionPassed($record, (array) $where['conditions'], $where['conjunction']);
      });
    }
    return $stmt;
  }

  /**
   * Apply the order by to the statement.
   *
   * @param \League\Csv\Statement $stmt
   *   The statement to apply the order by to.
   *
   * @return \League\Csv\Statement
   *   The statement with the order by applied.
   */
  protected function applyOrderBy(Statement $stmt): Statement {
    if ($order_by = $this->getObjectItem('order_by')) {
      foreach ($order_by as $sort) {
        $stmt = $stmt->orderBy(function (array $a, array $b) use ($sort): int {
          $direction = $sort->direction;
          return match ($direction) {
            self::DIRECTION_ASC => strnatcasecmp($a[$sort->column] ?? '', $b[$sort->column] ?? ''),
            self::DIRECTION_DESC => -strnatcasecmp($a[$sort->column] ?? '', $b[$sort->column] ?? ''),
            default => 0
          };
        });
      }
      return $stmt;
    }
    return $stmt;
  }

  /**
   * Apply the group by and aggregation.
   *
   * @param \League\Csv\TabularDataReader $result_set
   *   The result set to apply the group by and aggregation to.
   * @param array $group_by_columns
   *   The names of the columns to group by.
   * @param array $aggregated_columns
   *   The names of the columns to aggregate.
   *
   * @return \League\Csv\TabularDataReader
   *   The result set with the group by and aggregation applied.
   */
  protected function applyGroupBy(TabularDataReader $result_set, array $group_by_columns, array $aggregated_columns): TabularDataReader {
    // Ensure that all the columns that the grouping is going to apply to
    // existing columns.
    if ($group_by_columns !== array_intersect($group_by_columns, $result_set->getHeader())) {
      return $result_set;
    }

    $result_set->each(function ($row) use ($group_by_columns, $aggregated_columns) {
      $grouped_by_values = static::extractGroupedColumnsValuesFromRecord($row, $group_by_columns);
      static::processGroupByAndAggregation($grouped_by_values, $group_by_columns, $aggregated_columns, $row);
    });

    return $result_set;
  }

  /**
   * Add a column to the fields array.
   *
   * @param \stdClass $column
   *   The column to add.
   *
   * @return $this
   *   The current object.
   */
  protected function addColumn(\stdClass $column): static {
    $fields = $this->getObjectItem('fields');
    $field_added = FALSE;
    foreach ($fields as &$existing_field) {
      if ($existing_field->name === $column->name && !$existing_field->function !== $column->function) {
        $existing_field->function = $column->function;
        $field_added = TRUE;
        break;
      }
    }

    if (!$field_added) {
      $fields[] = $column;
    }
    return $this->setObjectItem('fields', $fields);
  }

  /**
   * Get an item from the object.
   *
   * @param string $item_key
   *   The key of the item to get.
   *
   * @return array
   *   The item.
   */
  protected function getObjectItem(string $item_key): array {
    $object = (object) json_decode($this->json, FALSE);
    return (array) ($object->{$item_key} ?? []);
  }

  /**
   * Set an item in the object.
   *
   * @param string $item_key
   *   The key of the item to set.
   * @param mixed $item
   *   The item to set.
   *
   * @return $this
   *   The current object.
   */
  protected function setObjectItem(string $item_key, mixed $item): static {
    $object = (object) json_decode($this->json, FALSE);
    $object->{$item_key} = $item;
    $this->json = json_encode($object);
    return $this;
  }

  /**
   * Build the conditions' array.
   *
   * @param array $condition
   *   The condition to build.
   *
   * @return array
   *   The built condition.
   */
  protected function buildConditions(array $condition): array {
    $new_condition = [];
    if ($condition['field'] instanceof ConditionInterface) {
      $conditions = $condition['field']->conditions();
      $new_condition['conjunction'] = $conditions['#conjunction'];
      unset($conditions['#conjunction']);
      $new_condition['conditions'] = [];
      foreach ($conditions as $cond) {
        $new_condition['conditions'][] = $this->buildConditions($cond);
      }
    }
    else {
      $new_condition['column'] = $condition['field'];
      $new_condition['value'] = $condition['value'];
      $new_condition['operator'] = $condition['operator'];
    }
    return $new_condition;
  }

  /**
   * Check if the condition is met.
   *
   * @param array $record
   *   The record to check the condition against.
   * @param array $conditions
   *   The conditions to check.
   * @param string $conjunction
   *   The conjunction to use to check the conditions.
   * @param bool $checked
   *   The initial value of the check.
   *
   * @return bool
   *   TRUE if the condition is met, FALSE otherwise.
   */
  protected static function conditionPassed(array $record, array $conditions, string $conjunction, bool $checked = TRUE): bool {
    if ($conjunction === 'AND' && !$checked) {
      return FALSE;
    }

    foreach ($conditions as $condition) {
      $condition = (array) $condition;
      if (isset($condition['conjunction'])) {
        // Running a recursive call to check nested conditions.
        $checked = static::conditionPassed($record, $condition['conditions'], $conjunction, $checked);
      }
      elseif (!isset($record[$condition['column']])) {
        // The column in the condition is not in the record columns.
        $checked = TRUE;
      }
      else {
        $checked = static::verifyCondition($record[$condition['column']], $condition['value'], $condition['operator']);
      }

      // For AND conjunction, we shouldn't bail out as soon as we have a
      // negative check.
      if ($conjunction === 'AND' && !$checked) {
        return FALSE;
      }
      // For OR conjunction, we should also bail out as soon as we have a
      // positive check.
      if ($conjunction === 'OR' && $checked) {
        return TRUE;
      }
    }

    return $checked;
  }

  /**
   * Verify the condition.
   *
   * @param mixed $record_value
   *   The value of the record.
   * @param mixed $condition_value
   *   The value of the condition.
   * @param string $operator
   *   The operator to use to compare the values.
   *
   * @return bool
   *   TRUE if the condition is met, FALSE otherwise.
   */
  protected static function verifyCondition(mixed $record_value, mixed $condition_value, string $operator): bool {
    // @todo remember to introduce case insensitive checks.
    return match($operator) {
      '=' => $record_value == $condition_value,
      '!=' => $record_value != $condition_value,
      'contains' => stripos($record_value, $condition_value) !== FALSE,
      'starts' => str_starts_with($record_value, $condition_value),
      'not_starts' => !str_starts_with($record_value, $condition_value),
      'ends' => !(strlen($condition_value) > 0) || str_ends_with($record_value, $condition_value),
      'not_ends' => !str_ends_with($record_value, $condition_value),
      'not' => stripos($record_value, $condition_value) === FALSE,
      'shorterthan' => strlen($record_value) < $condition_value,
      'longerthan' => strlen($record_value) > $condition_value,
      'regular_expression' => preg_match($condition_value, $record_value) === 1,
      // They gave us an operator we don't support.
      default => TRUE,
    };
  }

  /**
   * Extract the values of the grouped columns from the record.
   *
   * @param array $record
   *   The record to extract the values from.
   * @param array $grouped_columns
   *   The names of the grouped columns.
   *
   * @return array
   *   The values of the grouped columns.
   */
  protected static function extractGroupedColumnsValuesFromRecord(array $record, array $grouped_columns): array {
    $values = [];
    foreach ($grouped_columns as $column) {
      $values[] = $record[$column];
    }
    return $values;
  }

  /**
   * Get the selected fields.
   *
   * @return array
   *   The selected fields.
   */
  protected function getSelectedFields(): array {
    $fields = [];
    foreach ($this->getObjectItem('fields') as $field) {
      $fields[$field->name] = $field->function;
    }
    return $fields;
  }

  /**
   * Build a CSV reader from the grouped and aggregated records.
   *
   * @param array $options
   *   The options to use to build the CSV reader.
   *
   * @return \League\Csv\Reader
   *   The CSV reader.
   *
   * @throws \League\Csv\Exception
   *   If the CSV reader cannot be built.
   */
  protected function buildCsvReader(array $options): Reader {
    $records = [];
    $headers = [];
    foreach ($this->groupedAndAggregatedRecords as $record) {
      if (!$headers) {
        $headers = array_keys($record);
      }
      $records[] = array_values($record);
    }

    // The header at the beginning of the records.
    array_unshift($records, $headers);

    $tmp = new \SplTempFileObject();
    foreach ($records as $record) {
      $tmp->fputcsv($record);
    }
    $csv = Reader::createFromFileObject($tmp);

    $csv->setDelimiter($options['delimiter']);
    if (isset($options['header_index']) && $options['header_index'] !== NULL) {
      $csv->setHeaderOffset($options['header_index']);
    }
    return $csv;
  }

  /**
   * Process the group by and aggregation.
   *
   * @param array $group_by_values
   *   The values of the group by columns.
   * @param array $group_by_columns
   *   The names of the group by columns.
   * @param array $aggregated_columns
   *   The names of the aggregated columns.
   * @param array $row
   *   The row to process.
   */
  protected function processGroupByAndAggregation(array $group_by_values, array $group_by_columns, array $aggregated_columns, array $row): void {
    $key = implode('_____', $group_by_values);
    // Adding the grouping fields as part of the array before any aggregation.
    if (!isset($this->groupedAndAggregatedRecords[$key])) {
      $this->groupedAndAggregatedRecords[$key] = array_combine($group_by_columns, $group_by_values);
    }

    // Calculate aggregation if any.
    foreach ($aggregated_columns as $column => $function) {
      if ($function === 'count') {
        $counter = $this->groupedAndAggregatedRecords[$key][$column] ?? 0;
        $this->groupedAndAggregatedRecords[$key][$column] = $counter + (isset($row[$column]) ? 1 : 0);
      }
      elseif ($function === 'sum' && is_numeric($row[$column])) {
        $sum = $this->groupedAndAggregatedRecords[$key][$column] ?? 0;
        $this->groupedAndAggregatedRecords[$key][$column] = $sum + $row[$column];
      }
      elseif ($function === 'min' && is_numeric($row[$column])) {
        $min = $this->groupedAndAggregatedRecords[$key][$column] ?? $row[$column];
        $this->groupedAndAggregatedRecords[$key][$column] = min($min, $row[$column]);
      }
      elseif ($function === 'max' && is_numeric($row[$column])) {
        $max = $this->groupedAndAggregatedRecords[$key][$column] ?? $row[$column];
        $this->groupedAndAggregatedRecords[$key][$column] = max($max, $row[$column]);
      }
      elseif ($function === 'avg' && is_numeric($row[$column])) {
        $index_sum = $column . '_sum';
        $index_count = $column . '_count';
        $sum = $this->groupedAndAggregatedRecords[$key][$index_sum] ?? 0;
        $count = $this->groupedAndAggregatedRecords[$key][$index_count] ?? 0;
        $sum += $row[$column];
        $count += 1;
        $this->groupedAndAggregatedRecords[$key][$index_sum] = $sum;
        $this->groupedAndAggregatedRecords[$key][$index_count] = $count;
        $this->groupedAndAggregatedRecords[$key][$column] = $sum / $count;
      }
    }
  }

}
